#include <stdio.h>
#include <stdlib.h>

#include "vcl.h"
#include "vrt.h"

#include "vcc_if.h"
#include "parse_encoding.h"
#include "wb.h"

#define SKIP_EMPTY(next, ap)				\
	do {						\
		next = va_arg((ap), const char *);	\
	} while ((next) == NULL || *(next) == '\0')

/*
 * SAFE LENGTH ESTIMATES
 */

#define base64_decode_l(l)	((l * 3 / 4) + 1)
#define base64_encode_l(l)	(((((l * 4 / 3) + 3) / 4) * 4) + 1)
#define hex_encode_l(l)		((l * 2) + 1)
#define hex_decode_l(l)		((l / 2) + 1)

static const struct vmod_priv const null_blob[1] =
{
	{
		.priv = "",
		.len = 0,
		.free = NULL
	}
};

struct b64_state_s {
	uint32_t	magic;
#define B64_STATE_MAGIC	0xc0b64b64
	unsigned char	bytes[2];
	ssize_t	l;
};

static struct b64_alphabet {
	char *b64;
	char i64[256];
	int padding;
} b64_alphabet[__MAX_ENCODING];

static const char hex_alphabet[][16] = {
	"0123456789abcdef",
	"0123456789ABCDEF"
};

static void
alpha_init(struct b64_alphabet *alpha)
{
	int i;
	const char *p;

	for (i = 0; i < 256; i++)
		alpha->i64[i] = -1;
	for (p = alpha->b64, i = 0; *p; p++, i++)
		alpha->i64[(int)*p] = (char)i;
	if (alpha->padding)
		alpha->i64[alpha->padding] = 0;
}

int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e event)
{
	(void) priv;

	if (event != VCL_EVENT_LOAD)
		return 0;

	b64_alphabet[BASE64].b64 =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		"ghijklmnopqrstuvwxyz0123456789+/";
	b64_alphabet[BASE64].padding = '=';
	b64_alphabet[BASE64URL].b64 =
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		 "ghijklmnopqrstuvwxyz0123456789-_";
	b64_alphabet[BASE64URL].padding = '=';
	b64_alphabet[BASE64URLNOPAD].b64 =
		 "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef"
		 "ghijklmnopqrstuvwxyz0123456789-_";
	b64_alphabet[BASE64URLNOPAD].padding = 0;
	alpha_init(&b64_alphabet[BASE64]);
	alpha_init(&b64_alphabet[BASE64URL]);
	alpha_init(&b64_alphabet[BASE64URLNOPAD]);
	return (0);
}

static ssize_t
id_encode(const enum encoding enc, char *restrict const buf,
	  const size_t maxlen, const char *restrict const in,
	  const size_t inlen)
{
	(void) enc;
	AN(buf);

	if (maxlen < inlen + 1)
		return -1;
	if (in == NULL || inlen == 0)
		return 0;

	memcpy(buf, in, inlen);
	return inlen;
}

static ssize_t
id_decode(const enum encoding enc, char *restrict const buf,
	  const size_t maxlen, const char *restrict const p, va_list ap)
{
	const char *next;
	char *dest = buf;
	size_t len, outlen = 0;

	(void) enc;
	AN(buf);

	if (p == vrt_magic_string_end)
		return 0;

	SKIP_EMPTY(next, ap);
	if (next == vrt_magic_string_end && (p == NULL || *p == '\0'))
		return 0;

	if (p != NULL && *p != '\0') {
		if ((len = strlen(p)) > maxlen)
			return -1;
		memcpy(buf, p, len);
		outlen = len;
		dest += len;
	}

	while (next != vrt_magic_string_end) {
		len = strlen(next);
		if ((outlen += len) > maxlen)
			return -1;
		memcpy(dest, next, len);
		dest += len;
		SKIP_EMPTY(next, ap);
	}

	return outlen;
}

/*
 * Base64-encode *in (size: inlen) into the blob supplied as rblob. If
 * there is insufficient space, it will bail out and return
 * NULL. Otherwise, it will null-terminate and return the blob.
 * The enc argument specifies the alphabet and whether padding is used.
 * Inspired heavily by gnulib/Simon Josefsson (as referenced in RFC4648)
 *
 * XXX: tmp[] and idx are used to ensure the reader (and author) retains
 * XXX: a limited amount of sanity. They are strictly speaking not
 * XXX: necessary, if you don't mind going crazy.
 */
static ssize_t
base64_encode(const enum encoding enc, char *restrict const buf,
	      const size_t maxlen, const char *restrict const inbuf,
	      const size_t inlength)
{
	struct b64_alphabet *alpha = &b64_alphabet[enc];
	char *p = buf;
	const char *in = inbuf;
	size_t outlen = maxlen, inlen = inlength;
	unsigned char tmp[3], idx;

	AN(buf);
	if (in == NULL || inlen == 0)
		return 0;
	if (outlen < base64_encode_l(inlen))
		return -1;

	while (1) {
		assert(inlen);
		assert(outlen > 3);

		tmp[0] = (unsigned char) in[0];
		tmp[1] = (unsigned char) in[1];
		tmp[2] = (unsigned char) in[2];

		*p++ = alpha->b64[(tmp[0] >> 2) & 0x3f];

		idx = (tmp[0] << 4);
		if (inlen > 1)
			idx += (tmp[1] >> 4);
		idx &= 0x3f;
		*p++ = alpha->b64[idx];

		if (inlen > 1) {
			idx = (tmp[1] << 2);
			if (inlen > 2)
				idx += tmp[2] >> 6;
			idx &= 0x3f;

			*p++ = alpha->b64[idx];
		} else {
			if (alpha->padding)
				*p++ = alpha->padding;
		}

		if (inlen > 2) {
			*p++ = alpha->b64[tmp[2] & 0x3f];
		} else {
			if (alpha->padding)
				*p++ = alpha->padding;
		}

		/*
		 * XXX: Only consume 4 bytes, but since we need a fifth for
		 * XXX: NULL later on, we might as well test here.
		 */
		assert(outlen >= 5);

		outlen -= 4;

		if (inlen < 4)
			break;

		inlen -= 3;
		in += 3;
	}

	assert(outlen > 0);
	return p - buf;
}

/*
 * en/decoders need to work for the string_list vargs case (implicit length) as
 * well as for the blob case (explicit length). Fortunately, the blob case is
 * not varadic, so we can assume just one length argument _or_ varags
 *
 * rblob: alloc additional whitespace so any result can also be (safely)
 *	  used as a string.
 */

static ssize_t
hex_encode(const enum encoding enc, char *restrict const buf,
	   const size_t maxlen, const char *restrict const in,
	   const size_t inlen)
{
	char *p = buf;
	const char *alphabet = hex_alphabet[0];

	AN(buf);
	if (in == NULL || inlen == 0)
		return 0;
	if (maxlen < hex_encode_l(inlen))
		return -1;

	if (enc != HEXLC)
		alphabet = hex_alphabet[1];

	for (int i = 0; i < inlen; i++) {
		*p++ = alphabet[(in[i] & 0xf0) >> 4];
		*p++ = alphabet[in[i] & 0x0f];
	}

	return p - buf;
}

static inline VCL_STRING
concat_va(struct ws *ws, const char *p, va_list ap)
{
	const char *next;
	struct wb_s wb;

	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
	if (p == vrt_magic_string_end)
		return "";

	SKIP_EMPTY(next, ap);

	if (next == vrt_magic_string_end) {
		if ( p == NULL || *p == '\0')
			return "";
		else
			return p;
	}

	if (!wb_create(ws, &wb) || !wb_append(&wb, p, -1))
		return NULL;

	do {
		if (!wb_append(&wb, next, -1)) {
			wb_reset(&wb);
			return NULL;
		}
		SKIP_EMPTY(next, ap);
	} while (next != vrt_magic_string_end);

	if (wb_len(&wb) == 0) {
		wb_reset(&wb);
		return "";
	}
	return wb_finish(&wb, NULL);
}

static size_t
decode_l_va(enum encoding dec, const char * const p, va_list ap)
{
	size_t len = 0;
	const char *next;

	if (p == vrt_magic_string_end)
		return 0;
	if (p != NULL)
		len = strlen(p);

	while ((next = va_arg(ap, const char *)) != vrt_magic_string_end)
		if (next != NULL)
			len += strlen(next);

	switch(dec) {
	case IDENTITY:
		return len;
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_decode_l(len);
	case HEX:
	case HEXLC:
		return hex_decode_l(len);
	default:
		WRONG("Illegal decoding");
	}
}

/* Functions */

VCL_BLOB
vmod_decode(VRT_CTX, VCL_ENUM decs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	va_list ap;
	struct wb_s wb;
	struct vmod_priv *b;
	char *snap, *buf;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	assert(dec != _INVALID);
	assert(dec == IDENTITY); // XXX

	snap = WS_Snapshot(ctx->ws);
	if ((b = WS_Alloc(ctx->ws, sizeof(struct vmod_priv))) == NULL)
		return NULL;
	if (wb_create(ctx->ws, &wb) == NULL) {
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	buf = wb.w;

	va_start(ap, p);
	len = id_decode(dec, buf, wb_space(&wb), p, ap);
	va_end(ap);

	if (len == -1) {
		wb_reset(&wb);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		WS_Reset(ctx->ws, snap);
		return null_blob;
	}
	wb_advance(&wb, len);
	WS_ReleaseP(ctx->ws, wb.w);
	b->priv = buf;
	b->len = len;
	b->free = NULL;
	return b;
}

VCL_STRING
vmod_encode(VRT_CTX, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding enc = parse_encoding(encs);
	struct wb_s wb;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(b);

	if (wb_create(ctx->ws, &wb) == NULL)
		return NULL;

	switch(enc) {
	case IDENTITY:
		len = id_encode(enc, wb.w, wb_space(&wb), b->priv, b->len);
		break;
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		len = base64_encode(enc, wb.w, wb_space(&wb), b->priv, b->len);
		break;
	case HEX:
	case HEXLC:
		len = hex_encode(enc, wb.w, wb_space(&wb), b->priv, b->len);
		break;
	default:
		WRONG("Illegal encoding");
	}

	if (len == -1) {
		wb_reset(&wb);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		return "";
	}
	wb_advance(&wb, len);
	return wb_finish(&wb, NULL);
}

VCL_STRING __match_proto__(td_convert_transcode)
vmod_transcode(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	struct wb_s wb;
	ssize_t len;
	va_list ap;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	/*
	 * If the encoding and decoding are the same, just return the
	 * concatenated string.
	 * XXX: Doesn't reject illegal decodings.
	 */
	if (enc == dec) {
		va_start(ap, p);
		VCL_STRING s = concat_va(ctx->ws, p, ap);
		va_end(ap);
		return s;
	}

	/* Allocate space for the decoded blob on the stack */
	va_start(ap, p);
	size_t l = decode_l_va(dec, p, ap);
	va_end(ap);
	if (l == 0)
		return "";
	assert(l > 0);
	char buf[l];

	va_start(ap, p);
	switch(dec) {
	case IDENTITY:
		len = id_decode(dec, buf, l, p, ap);
		break;
	default:
		WRONG("Illegal decoding");
	}
	va_end(ap);

	if (len == -1)
		return NULL;

	if (wb_create(ctx->ws, &wb) == NULL)
		return NULL;

	switch(enc) {
	case IDENTITY:
		len = id_encode(enc, wb.w, wb_space(&wb), buf, len);
		break;
	case HEX:
	case HEXLC:
		len = hex_encode(enc, wb.w, wb_space(&wb), buf, len);
		break;
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		len = base64_encode(enc, wb.w, wb_space(&wb), buf, len);
		break;
	default:
		WRONG("Illegal encoding");
	}

	if (len == -1) {
		wb_reset(&wb);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		return "";
	}
	wb_advance(&wb, len);
	return wb_finish(&wb, NULL);
}
